from s4sdk.resource.abc import Resource
from s4sdk.metadata import ResourceType


class BinaryResource(Resource):
    def __init__(self, type: ResourceType, group: int, instance: int, content: bytes):
        super().__init__(type=type, group=group, instance=instance, content=content)

    @classmethod
    def read(cls, path: str, **kwargs):
        file = open(path, "rb")
        instance = cls(
            type=kwargs.get("type"),
            group=kwargs.get("group"),
            instance=kwargs.get("instance"),
            content=file.read()
        )
        return instance

    @property
    def content(self) -> bytes:
        return self._bstr

    def write(self, path: str):
        with open(path, "wb") as file:
            file.write(self._bstr)




import weakref
import io
import contextlib


class FormatException(Exception):
    pass


class WeakIdDict(dict):
    # This is completely untested
    def __init__(self):
        super().__init__()

    def __getitem__(self, key):
        try:
            obj, val = super().__getitem__(id(key))
            if obj() is not key:
                # SHould not happen; keys are dropped by the finalizer
                super().__delitem__[id(key)]
                raise KeyError(key)
            return val
        except KeyError:
            raise KeyError(key)

    def __setitem__(self, key, val):
        keyid = id(key)

        def cb():
            super().__delitem__(keyid)
        keyref = weakref.ref(key, cb)
        super().__setitem__(keyid, (keyref, val))
        del key # Drop key from locals so that we don't hold on to a
                # reference in the lambda's closure

    def __delitem__(self, key):
        super().__delitem__(id(key))

    def __contains__(self, key):
        if super().__contains__(id(key)):
            return False
        if super().__getitem__(id(key))[0]() is not key:
            return False
        return True


class LazyProperty:
    """A property that lazily evaluates and caches its result."""
    def __init__(self, thunk):
        self.thunk = thunk

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            res = self.thunk(instance)
            setattr(instance, self.thunk.__name__, res)
            return res

    def __del__(self, instance, owner):
        if instance is not None:
            pass


class Thunk:
    """A lazily-evaluated value"""
    def __init__(self, thunk):
        self.thunk = thunk
        self.setp = False
        self._value = None

    @property
    def value(self):
        if not self.setp:
            self._value = self.thunk()
            self.setp = True
        return self._value


class BinPacker:
    writable = False

    def __init__(self, bstr, mode="r"):
        if mode == 'w':
            self.writable = True
        if isinstance(bstr, bytes):
            self.raw_len = len(bstr)
            bstr = io.BytesIO(bstr)
        else:
            bstr.seek(0, io.SEEK_END)
            self.raw_len = bstr.tell()
            bstr.seek(0)
        self.raw = bstr

    def close(self):
        self.raw.close()

    @property
    def off(self):
        return self.raw.tell()

    @off.setter
    def off(self, val):
        if val <= self.raw_len or self.writable:
            self.raw.seek(val)
        else:
            raise ValueError("Seek off of end of file")

    @contextlib.contextmanager
    def at(self, posn):
        """Temporarily read from another place. If posn is None, just save the
        current position.

        This is intended to be used like:
        with reader.at(offset):
           # read some stuff

        """
        saved = self.off
        try:
            if posn is not None:
                self.off = posn
            yield
        finally:
            self.off = saved

    def put_raw_bytes(self, bstr):
        self.raw.write(bstr)

    def _put_int(self, i, len, signedp):
        self.put_raw_bytes(i.to_bytes(len, "little", signed=signedp))

    def put_int8(self, i):   self._put_int(i, 1, True)
    def put_int16(self, i):  self._put_int(i, 2, True)
    def put_int32(self, i):  self._put_int(i, 4, True)
    def put_int64(self, i):  self._put_int(i, 8, True)

    def put_uint8(self, i):  self._put_int(i, 1, False)
    def put_uint16(self, i): self._put_int(i, 2, False)
    def put_uint32(self, i): self._put_int(i, 4, False)
    def put_uint64(self, i): self._put_int(i, 8, False)

    def put_strz(self, s):
        self.put_raw_bytes(s.encode('utf-8'))
        s.put_raw_bytes(b'\0')

    # Reading methods
    def get_raw_bytes(self, count):
        return self.raw.read(count)

    def get_off32(self):
        """Read an offset relative to the current position; returns the absolute offset"""
        off = self.off
        res = self.get_int32()
        if res == -0x80000000:
            return None
        else:
            return off + res

    def _get_int(self, size, signed=False):
        return int.from_bytes(self.get_raw_bytes(size),
                              "little", signed=signed)

    def get_uint64(self): return self._get_int(8, False)
    def get_int64(self): return self._get_int(8, True)

    def get_uint32(self): return self._get_int(4, False)
    def get_int32(self): return self._get_int(4, True)

    def get_uint16(self): return self._get_int(2, False)
    def get_int16(self): return self._get_int(2, True)

    def get_uint8(self): return self._get_int(1, False)
    def get_int8(self): return self._get_int(1, True)


    def get_string(self):
        """Read a null-terminated string"""
        res_queue = []
        while True:
            pos = self.off
            block = self.get_raw_bytes(16)
            if len(block) == 0:
                raise FormatException("Unexpected EOF")
            zpos = block.find(b'\0')
            if zpos != -1:
                self.off = pos + zpos + 1 # +1 to seek past null byte
                block = block[0:zpos]
                res_queue.append(block)
                break
            else:
                res_queue.append(block)
        return b''.join(res_queue)

    def get_relstring(self):
        """Read a string from the next offset, read as an off32"""
        off = self.get_off32()
        if off is not None:
            with self.at(off):
                return self.get_string()
        else:
            return None

    def align(self, size):
        """Align input position to a multiple of size"""
        off = (self.off + size - 1)
        self.off = off - (off % size)



# Provides useful tools for working with packages, including metapackage support
import os.path

from s4sdk import utils
from s4sdk.package.metapackage import MetaPackage
from s4sdk.package.dbpf import DbpfPackage
from s4sdk.package.dirpackage import DirPackage
from s4sdk import metadata
from s4sdk.resource.abc import Resource
from s4sdk.resource import Resource as _Resource, ResourceID, ResourceFilter
from s4sdk.resource.stbl import StringTable, StringTableMetadata
from s4sdk.resource.binary import BinaryResource
import pandas as pd
from typing import List


def open_package(filename, mode="r"):
    absname = os.path.abspath(filename)
    if mode == "r":
        if not os.path.exists(filename):
            raise FileNotFoundError(
                "No such file or directory: %s" % (filename,))
        if os.path.isdir(filename):
            return DirPackage(absname)
        with open(filename, "rb") as f:
            magic = f.read(4)
            if magic == b"DBPF":
                return DbpfPackage(filename)
        if filename.lower().endswith(".meta"):
            # It's a metapackage...
            try:
                return MetaPackage.open(filename)
            except UnicodeError:
                raise utils.FormatException("Invalid unicode in metapackage")
        raise utils.FormatException("Couldn't identify package format")
    elif mode == 'w':
        if filename.endswith(".package"):
            return DbpfPackage(filename, "w")
        elif filename.endswith("/") or os.path.isdir(filename):
            return DirPackage(filename, mode="w")


class Package:
    def __init__(self, dbfile):
        self.dbfile = dbfile
        child_generator = self.dbfile.scan_index()
        self.instances = [instance for instance in child_generator]

    @classmethod
    def from_package(cls, path: str):
        return cls(dbfile=open_package(path, mode="r"))

    @classmethod
    def create_empty(cls, path: str):
        if os.path.exists(path):
            raise ValueError(f"Package file at {path} already exist")
        return cls(dbfile=open_package(path, mode="w"))

    def list(self):
        return pd.DataFrame.from_dict({
            idx: {
                "type": instance.type,
                "group": instance.group,
                "instance_id": instance.instance,
                "resource_key": instance
            }
            for idx, instance in enumerate(self.instances)
        }, orient="index")

    def get(self, instance_id: int | List[int]) -> Resource:
        if not isinstance(instance_id, list):
            instance_id = [instance_id]
        for instance in self.instances:
            if instance.instance in instance_id:
                instance_id.remove(instance.instance)
                _resource = self.dbfile[instance]
                if _resource.id.type == metadata.ResourceType.STBL.value:
                    return StringTable.read_bytes(bstr=_resource.content)
                else:
                    return BinaryResource(
                        type=metadata.classify_type(instance.type),
                        group=instance.group,
                        instance=instance.instance,
                        content=_resource.content
                    )
        if instance_id:
            raise ValueError(f"Cannot find these instance in the package: {instance_id}")

    def insert(self, resource: Resource | List[Resource]):
        if not isinstance(resource, list):
            resource = [resource]
        for rsrc in resource:
            self.dbfile.put(rid=rsrc.rid, content=rsrc.content)
            self.dbfile.commit()

    def remove(self, instance_id: int | List[int]):
        if not isinstance(instance_id, list):
            instance_id = [instance_id]
        pass

    def export(self, instance_id: int, path: str):
        res = self.get(instance_id=instance_id)
        res.write(path)



from enum import Enum


class ResourceType(Enum):
    STBL: int = 570775514
    GFX: int = 1659684250
    DDS: int = 11720834
    SIM_DATA: int = 1415235194
    OBJ_TUNING: int = 3055412916


def classify_type(type: int):
    for type_str in ResourceType.__annotations__.keys():
        if ResourceType[type_str].value == type:
            return ResourceType[type_str]
    raise ValueError(f"Resource type '{type}' is unsupported")



from collections import namedtuple
import re
import yaml


class Resource(namedtuple("Resource", 'id locator size package')):
    # Locator is a package-specific tuple that provides all necessary
    # info for the package object to read the content.
    @property
    def content(self):
        return self.package._get_content(self)

    def __eq__(self, other):
        return (self.id == other.id
                and self.locator == other.locator
                and self.size == other.size
                and self.package is other.package)

    def __ne__(self, other):
        return not (self == other)


class ResourceID(namedtuple("ResourceID", "group instance type")):
    """A DBPF resource ID. This can be used directly as a RID filter; it matches only itself"""
    __slots__ = ()

    DEFAULT_FMT = "maxis"

    FORMATTERS = {
        's4pe': 'S4_{id.type:08X}_{id.group:08X}_{id.instance:016X}',
        'colon': '{id.group:08x}:{id.instance:016x}:{id.type:08x}',
        'maxis': '{id.group:08x}!{id.instance:016x}.{id.type:08x}',
    }

    PARSERS = {
        's4pe': '^S4_{type}_{group}_{instance}(?:%%.*)?$',
        'colon': '^{group}:{instance}:{type}$',
        'maxis': '^{group}!{instance}.{type}$',
    }

    for fmt in PARSERS:
        PARSERS[fmt] = re.compile(PARSERS[fmt].format(
            type="(?P<type>[0-9A-Fa-f]{,8})",
            group="(?P<group>[0-9A-Fa-f]{,8})",
            instance="(?P<instance>[0-9A-Fa-f]{,16})"))

    def __str__(self):
        return self.FORMATTERS[self.DEFAULT_FMT].format(id=self)

    def match(self, rid):
        """Determine if this RID is the same as the argument. Strictly
        speaking, this is the same as __eq__, but ResourceFilters must
        implement this function
        """
        return rid == self

    def as_filename(self):
        return str(self)

    @classmethod
    def from_string(cls, string):
        # Handles:
        # - colon-format (group:instance:type)
        # - Maxis format (group!instance.type)
        # - S4 format (S4_type_group_instance(?:%%.*)?)
        for parser in cls.PARSERS.values():
            m = parser.match(string)
            if m:
                return cls(
                    int(m.group('group'), 16),
                    int(m.group('instance'), 16),
                    int(m.group('type'), 16),
                )
        else:
            raise ValueError("Invalid rid %s" % (string,))


def _represent_RID(dumper, rid):
    return dumper.represent_scalar('!s4/rid', str(rid))


yaml.add_representer(ResourceID, _represent_RID)


class ResourceFilter:
    """A simple resource filter; this matches iff all specified RID
    components are equal to the ones in the RID being tested"""

    __slots__ = ('group','instance','type')

    def __init__(self, group=None,instance=None,type=None):
        self.group = group
        self.instance = instance
        self.type = type

    def match(self, rid):
        return ((self.group is None or self.group == rid.group) and
                (self.instance is None or self.instance == rid.instance) and
                (self.type is None or self.type == rid.type))

    def __str__(self):
        group = ("%08x"%self.group) if self.group is not None else ""
        instance = ("%16x"%self.instance) if self.instance is not None else ""
        type = ("%08x"%self.type) if self.type is not None else ""
        return ":".join((group,instance,type))



from s4sdk import metadata
from abc import ABC, abstractmethod
from s4sdk.resource import ResourceID


class Resource(ABC):
    def __init__(self, type: metadata.ResourceType, content: bytes = None, group: int = 0, instance: int = None):
        self._group = group
        self._instance = instance
        self._bstr = content
        self._type = type.value

    @classmethod
    @abstractmethod
    def read(cls, path: str):
        pass

    @property
    def rid(self) -> ResourceID:
        return ResourceID(group=self.group, instance=self.instance, type=self.type)

    @property
    def group(self) -> int:
        return self._group

    @property
    def instance(self) -> int:
        return self._instance

    @property
    def type(self) -> metadata.ResourceType:
        return self._type

    @property
    @abstractmethod
    def content(self) -> bytes:
        return self._bstr

    @abstractmethod
    def write(self, path: str):
        pass



import abc
from s4sdk import resource


class AbstractPackage(metaclass=abc.ABCMeta):

    def __init__(self):
        self.__stbl_cache = None

    @abc.abstractmethod
    def scan_index(self, filter=None):
        pass

    @abc.abstractmethod
    def _get_content(self, resource):
        """Retrieve the content from Resource (which is guaranteed to have
        been returned from a __getitem__ call to this package)

        """

    @abc.abstractmethod
    def __getitem__(self, item):
        """Maps from a ResourceID to a Resource. Any other usage is an error.

        This method MUST NOT use any caches that are flushed by
        flush_index_cache.

        """

    def flush_index_cache(self):
        """Flush the index cache to save memory. This method is optional; some
        database formats may not need an index cache due to use of an
        efficient file format."""

    @property
    def stbl(self):
        if self.__stbl_cache is None:
            self.__stbl_cache = {}
            for stblid in self.scan_index(
                    resource.ResourceFilter(type=0x220557DA)):
                for key, value in stbl.read_stbl(self[stblid]):
                    self.__stbl_cache[key] = value
        return self.__stbl_cache

    def close(self):
        """Close the package, freeing any OS-level resources if necessary"""
        self.commit()

    def commit(self):
        """Commit any outstanding changes to disk without closing the
        file. Does not need to be called for read-only package access.

        """



import os.path
from collections import namedtuple

from .abstractpackage import AbstractPackage
from .. import resource


class FileLocator(namedtuple("FileLocator", "filename")):
    pass


class DirPackage(AbstractPackage):
    """A dirpackage is the most versatile form of package: it is the only
    form that can be opened read/write. It is simply a loose
    collection of properly-named files (in Maxis, S4pe, or colon
    format) in a directory. By default, it writes files in Maxis
    format. """

    def __init__(self, path, *args, mode="r", config=None, **kwargs):
        """The config file is completely overridden by any config file that
        already exists in the directory.
        """
        super().__init__()

        self.path = os.path.abspath(path)
        if mode == "r":
            if not os.path.exists(self.path):
                raise FileNotFoundError(
                    "Couldn't open directory package at %s"% (path,))
            self._index_cache = None
            self.writable = False
        elif mode == "w":
            if not os.path.isdir(self.path):
                os.makedirs(self.path)
            self._index_cache = None
            self.writable = True

    @property
    def _index(self):
        if self._index_cache is not None:
            return self._index_cache
        self._index_cache = {}
        for fname in os.listdir(self.path):
            fullpath = os.path.join(self.path, fname)
            if not os.path.isfile(fullpath):
                continue
            try:
                rid = resource.ResourceID.from_string(fname)
            except ValueError:
                # Ignore the file
                pass
            else:
                self._index_cache[rid] = resource.Resource(
                    id=rid,
                    locator = fullpath,
                    size = os.stat(fullpath).st_size,
                    package=self)
        return self._index_cache

    def scan_index(self, filter=None):
        for x in self._index:
            if filter is None or filter.match(x):
                yield x

    def _get_content(self, resource):
        return open(resource.locator, "rb").read()
    def __getitem__(self, rid):
        return self._index[rid]
    def flush_index_cache(self):
        self._index_cache = None

    def put(self, rid, value):
        fname = os.path.join(self.path, rid.as_filename())
        with open(fname, "wb") as f:
            f.write(value)
        self._index[rid] = resource.Resource(
            id=rid,
            locator=fname,
            size=len(value),
            package=self)



import pandas as pd

from s4sdk.resource.abc import Resource
from s4sdk import utils, metadata
from dataclasses import dataclass
from collections import defaultdict


@dataclass
class StringTableMetadata:
    magic: bytes
    version: int
    compressed: int
    num_entries: int
    str_length: int

    @classmethod
    def from_binary_pack(cls, b_pack: utils.BinPacker):
        magic = b_pack.get_raw_bytes(4)
        if magic != b'STBL':
            raise utils.FormatException("Bad magic")
        version = b_pack.get_uint16()
        if version != 5:
            raise utils.FormatException("We only support STBLv5")
        compressed = b_pack.get_int8()
        num_entries = b_pack.get_uint64()
        b_pack.off += 2
        str_length = b_pack.get_uint32()
        return cls(magic=magic, version=version, compressed=compressed, num_entries=num_entries, str_length=str_length)

    @classmethod
    def from_empty(cls, num_entries: int, str_length: int):
        return cls(magic=b'STBL', version=5, compressed=0, num_entries=num_entries, str_length=str_length)


class StringTable(Resource):
    def __init__(self, meta_data=None, entries=None):
        super().__init__(type=metadata.ResourceType.STBL)
        self.meta_data = meta_data or StringTableMetadata.from_empty(num_entries=0, str_length=0)
        self.entries: pd.DataFrame = entries if entries is not None else pd.DataFrame()

    @classmethod
    def read(cls, path: str):
        bstr = open(path, "rb").read()
        return StringTable.read_bytes(bstr=bstr)

    @classmethod
    def read_bytes(cls, bstr: bytes):
        b_pack = utils.BinPacker(bstr)
        meta_data = StringTableMetadata.from_binary_pack(b_pack=b_pack)

        content = defaultdict(dict)
        for row in range(meta_data.num_entries):
            key_hash = b_pack.get_uint32()
            flags = b_pack.get_uint8()
            length = b_pack.get_uint16()
            val = b_pack.get_raw_bytes(length).decode('utf-8')
            content["key_hash"][row] = key_hash
            content["flags"][row] = flags
            content["length"][row] = length
            content["val"][row] = val
        entries = pd.DataFrame.from_dict(content)
        return cls(meta_data=meta_data, entries=entries)

    @classmethod
    def read_csv(cls, path: str):
        content = pd.read_csv(path).to_dict()
        n = len(content["key_hash"])
        total_len = 0
        str_len = 0
        for row in range(0, n):
            if content["val"][row] != content["val"][row]:
                new_len = 0
            else:
                new_len = len(content["val"][row].encode("utf-8"))
            content["length"][row] = new_len
            total_len += new_len
            str_len += len(content["val"][row]) if new_len else new_len

        num_entries = n
        str_len = total_len + num_entries

        meta_data = StringTableMetadata.from_empty(num_entries=num_entries, str_length=str_len)
        entries = content
        return cls(meta_data=meta_data, entries=entries)

    def write(self, path: str):
        b_pack = self.pack_bytes()
        with open(path, "wb") as file:
            file.write(b_pack.raw.getbuffer())
        b_pack.close()

    def write_csv(self, path: str):
        self.entries.to_csv(path, index=False)

    @property
    def content(self) -> bytes:
        return self.pack_bytes().raw

    def pack_bytes(self) -> utils.BinPacker:
        b_pack = utils.BinPacker(bstr=b'', mode='w')
        self._repack_metadata(b_pack=b_pack)
        self._repack_content(b_pack=b_pack)
        return b_pack

    def _repack_metadata(self, b_pack: utils.BinPacker):
        b_pack.put_raw_bytes(self.meta_data.magic)
        b_pack.put_uint16(self.meta_data.version)
        b_pack.put_uint8(self.meta_data.compressed)
        b_pack.put_uint64(self.meta_data.num_entries)
        b_pack.off += 2
        b_pack.put_uint32(self.meta_data.str_length)

    def _repack_content(self, b_pack: utils.BinPacker):
        contents = self.entries.to_dict(orient="index")
        for row, content in contents.items():
            b_pack.put_uint32(content.get("key_hash"))
            b_pack.put_uint8(content.get("flags"))
            b_pack.put_uint16(content.get("length"))
            val = content.get("val")
            b_pack.put_raw_bytes("".encode("utf-8") if val != val else val.encode("utf-8"))


def read_stbl(bstr):
    """Parse a string table (ID 0x220557DA)"""
    f = utils.BinPacker(bstr)
    if f.get_raw_bytes(4) != b'STBL':
        raise utils.FormatException("Bad magic")
    version = f.get_uint16()
    if version != 5:
        raise utils.FormatException("We only support STBLv5")
    compressed = f.get_uint8()
    numEntries = f.get_uint64()
    f.off += 2
    mnStringLength = f.get_uint32()
    # This is the total size of all the strings plus one null byte per string (to make the parsing code faster, probably)

    entries = {}
    size = 0
    for _ in range(numEntries):
        keyHash = f.get_uint32()
        flags = f.get_uint8() # What is in this? It's always 0.
        length = f.get_uint16()
        val = f.get_raw_bytes(length).decode('utf-8')
        yield keyHash, val



import io
from collections import namedtuple
import zlib

from s4sdk.package.abstractpackage import AbstractPackage
from s4sdk import resource, utils


class DbpfLocator(namedtuple("DbpfLocator", 'offset raw_len compression')):
    @property
    def deleted(self):
        return self.compression[0] == 0xFFE0


class _DbpfReader(utils.BinPacker):
    _Header = namedtuple('_Header',
                         'file_version user_version ctime ' +
                         'mtime index_count index_pos index_size')

    @property
    def header(self):
        if hasattr(self, "_header"):
            return self._header

        with self.at(0):
            if self.get_raw_bytes(4) != b'DBPF':
                raise utils.FormatException(
                    "Not a valid DBPF file; invalid magic")

            fileVersion = (self.get_uint32(), self.get_uint32())
            if fileVersion != (2, 1):
                raise utils.FormatException("Only DBPF v2.1 is supported")

            userVersion = (self.get_uint32(), self.get_uint32())
            _unused1 = self.get_uint32()
            mnCreationTime = self.get_uint32()
            mnUpdatedTime = self.get_uint32()
            _unused2 = self.get_uint32()
            indexRecordEntryCount = self.get_uint32()
            indexRecordPosLow = self.get_uint32()
            indexRecordSize = self.get_uint32()
            self.off += 16
            indexRecordPosHigh = self.get_uint32()
            self.off += 4 * 6

            if indexRecordPosHigh != 0:
                indexRecordPos = indexRecordPosHigh
            else:
                indexRecordPos = indexRecordPosLow
            self._header = self._Header(
                fileVersion,
                userVersion,
                mnCreationTime,
                mnUpdatedTime,
                indexRecordEntryCount,
                indexRecordPos,
                indexRecordSize
            )
            return self._header

    def get_index(self, package=None):
        # Package is used for the package field in Resource
        # This doesn't cache at all.
        _CONST_TYPE = 1
        _CONST_GROUP = 2
        _CONST_INSTAMCE_EX = 4
        header = self.header
        if header.index_pos == 0:
            if header.index_count != 0:
                raise utils.FormatException(
                    "Package contains entries but no index")
            return

        with self.at(header.index_pos):
            flags = self.get_uint32()
            if flags & _CONST_TYPE:        entry_type = self.get_uint32()
            if flags & _CONST_GROUP:       entry_group = self.get_uint32()
            if flags & _CONST_INSTAMCE_EX: entry_inst_ex = self.get_uint32()

            for _ in range(header.index_count):
                if not flags & _CONST_TYPE:
                    entry_type = self.get_uint32()
                if not flags & _CONST_GROUP:
                    entry_group = self.get_uint32()
                if not flags & _CONST_INSTAMCE_EX:
                    entry_inst_ex = self.get_uint32()
                entry_inst = self.get_uint32()
                entry_pos = self.get_uint32()
                entry_size = self.get_uint32()
                entry_size_decompressed = self.get_uint32()
                if entry_size & 0x80000000:
                    entry_compressed = (self.get_uint16(), self.get_uint16())
                else:
                    entry_compressed = (0, 1)
                entry_size &= 0x7FFFFFFF
                locator = DbpfLocator(entry_pos, entry_size, entry_compressed)
                with self.at(None):
                    yield resource.Resource(
                        resource.ResourceID(entry_group,
                                            entry_inst_ex << 32 | entry_inst,
                                            entry_type),
                        locator,
                        entry_size_decompressed,
                        package)


class _DbpfWriter:
    def __init__(self, fstream):
        self.f = utils.BinPacker(fstream, mode="w")
        # Skip over header. The official docs say the header is 92
        # bytes, but all the RE'd docs say 96. Treating an extra 4
        # bytes as reserved won't hurt, so we just use 96 here
        self.f.off = 96

    def put_rsrc(self, rid, content):
        off = self.f.off
        zcontent = zlib.compress(content)
        self.f.put_raw_bytes(zcontent)
        locator = DbpfLocator(off, len(zcontent), (0x5A42, 1))
        return locator

    def write_index(self, idx):
        with self.f.at(None):
            idx_start = self.f.off
            idx_count = 0
            # Save the current position, in case we decide to write
            # more content

            # For now, don't try to optimize the index by sharing
            # group/type/etc; it's fairly unlikely that it will be
            # possible to save a significant amount of space unless
            # the file is very small, in which case who cares?
            self.f.put_uint32(0)  # No flags

            for rsrc in idx.values():
                idx_count += 1
                self.f.put_uint32(rsrc.id.type)
                self.f.put_uint32(rsrc.id.group)
                self.f.put_uint32(rsrc.id.instance >> 32)
                self.f.put_uint32(rsrc.id.instance & 0xFFFFFFFF)
                self.f.put_uint32(rsrc.locator.offset)
                if rsrc.locator.raw_len & 0x80000000 != 0:
                    raise RuntimeError("File must be smaller than 2GB")
                # We always compress, so we always need the ExtendedCompression
                # bit set
                self.f.put_uint32(rsrc.locator.raw_len | 0x80000000)
                self.f.put_uint32(rsrc.size)
                self.f.put_uint16(rsrc.locator.compression[0])
                self.f.put_uint16(rsrc.locator.compression[1])
            idx_end = self.f.off
        header = _DbpfReader._Header((2, 1), (0, 0), 0, 0,
                                     idx_count, idx_start, idx_end - idx_start)
        self.put_header(header)

    def put_header(self, header):
        with self.f.at(0):
            self.f.put_raw_bytes(b'DBPF')
            self.f.put_uint32(header.file_version[0])
            self.f.put_uint32(header.file_version[1])
            self.f.put_uint32(header.user_version[0])
            self.f.put_uint32(header.user_version[1])
            self.f.put_uint32(0)
            self.f.put_uint32(header.ctime)
            self.f.put_uint32(header.mtime)
            self.f.put_uint32(0)
            self.f.put_uint32(header.index_count)
            self.f.put_uint32(0)
            self.f.put_uint32(header.index_size)
            self.f.put_uint32(0)
            self.f.put_uint32(0)
            self.f.put_uint32(0)
            self.f.put_uint32(3)
            self.f.put_uint32(header.index_pos)
            for _ in range(6):
                self.f.put_uint32(0)


class DbpfPackage(AbstractPackage):
    """A Sims4 DBPF file. This is the format in Sims4 packages, worlds, etc"""

    def __init__(self, name, mode="r"):
        super().__init__()
        if isinstance(name, io.RawIOBase):
            self.file = _DbpfReader(name)
        else:
            if mode == 'r':
                self.file = _DbpfReader(open(name, "rb"))
                self._index_cache = None
                self.writable = False
            elif mode == 'w':
                self.file = _DbpfWriter(open(name, "w+b"))
                self._index_cache = {}
                self.writable = True

    def scan_index(self, filter=None):
        if self._index_cache is None:
            self._index_cache = {}
            for item in self.file.get_index(self):
                if item.locator.deleted:
                    continue
                self._index_cache[item.id] = item

        for key in self._index_cache:
            if filter is None or filter.match(key):
                yield key

    def __getitem__(self, resource):
        if self._index_cache == None:
            next(self.scan_index())
        return self._index_cache[resource]

    def _get_content(self, item):
        assert isinstance(item, resource.Resource)
        assert item.package is self
        with self.file.at(item.locator.offset):
            ibuf = self.file.get_raw_bytes(item.locator.raw_len)

        if item.locator.compression[0] == 0:
            return ibuf  # uncompressed
        elif item.locator.compression[0] == 0xFFFE:
            # BUG: I'm guessing "streamable compression" is the same
            # as RefPack, with a limited buffer size. This may or may
            # not be true, and even if it is, I'd need to know the
            # size of the buffer to do anything sensible.
            return decodeRefPack(ibuf)
        elif item.locator.compression[0] == 0xFFFF:
            return decodeRefPack(ibuf)
        elif item.locator.compression[0] == 0x5A42:
            return zlib.decompress(ibuf, 15, item.size)

    def flush_index_cache(self):
        # If we're writable, the in-memory "cache" is actually the
        # *only* copy of the index, so it shouldn't be flushed.
        if not self.writable:
            self._index_cache = None

    def commit(self):
        if self.writable:
            self.file.write_index(self._index_cache)

    def put(self, rid, content):
        if self.writable:
            locator = self.file.put_rsrc(rid, content)
            self._index_cache[rid] = resource.Resource(
                rid, locator, len(content), self)
        else:
            raise TypeError("Not a writable package")

    def close(self):
        super().close()
        self.file.close()


def decodeRefPack(ibuf):
    """Decode the DBPF compression. ibuf must quack like a bytes"""
    # Based on http://simswiki.info/wiki.php?title=Sims_3:DBPF/Compression
    # Sims4 compression has the first two bytes swapped

    iptr = optr = 0
    flags = ibuf[0]
    if ibuf[1] != 0xFB:
        raise utils.FormatException("Invalid compressed data")
    iptr = 2
    osize = 0  # output size
    for _ in range(4 if flags & 0x80 else 3):
        osize = (osize << 8) | ibuf[iptr]
        iptr += 1

    obuf = bytearray(osize)
    while iptr < len(ibuf):
        numPlaintext = numToCopy = copyOffset = 0
        # Copyoffset is 0-indexed back from obuf[optr]
        # I.e., copyoffset=0 ==> copying starts at obuf[optr-1]

        # Read a control code
        cc0 = ibuf[iptr]
        iptr += 1
        if cc0 <= 0x7F:
            cc1 = ibuf[iptr]
            iptr += 1
            cc = (cc0, cc1)
            numPlaintext = cc0 & 0x03
            numToCopy = ((cc0 & 0x1C) >> 2) + 3
            copyOffset = ((cc0 & 0x60) << 3) + cc1
        elif cc0 <= 0xBF:
            cc1 = ibuf[iptr]
            iptr += 1
            cc2 = ibuf[iptr]
            iptr += 1
            cc = (cc0, cc1, cc2)
            numPlaintext = (cc1 & 0xC0) >> 6
            numToCopy = (cc0 & 0x3F) + 4
            copyOffset = ((cc1 & 0x3F) << 8) + cc2
        elif cc0 <= 0xDF:
            cc1 = ibuf[iptr]
            iptr += 1
            cc2 = ibuf[iptr]
            iptr += 1
            cc3 = ibuf[iptr]
            iptr += 1
            cc = (cc0, cc1, cc2, cc3)
            numPlaintext = cc0 & 0x03
            numToCopy = ((cc0 & 0x0C) << 6) + cc3 + 5
            copyOffset = ((cc0 & 0x10) << 12) + (cc1 << 8) + cc2
        elif cc0 <= 0xFB:
            cc = (cc0,)
            numPlaintext = ((cc0 & 0x1F) << 2) + 4
            numToCopy = 0
        else:
            cc = (cc0,)
            numPlaintext = cc0 & 3
            numToCopy = 0

        # Copy from source
        obuf[optr:optr + numPlaintext] = ibuf[iptr:iptr + numPlaintext]
        iptr += numPlaintext
        optr += numPlaintext

        # Copy from output
        for _ in range(numToCopy):
            obuf[optr] = obuf[optr - 1 - copyOffset]
            optr += 1
    # Done decompressing
    return bytes(obuf)



from .abstractpackage import AbstractPackage


class MetaPackage(AbstractPackage):
    """A stack of packages, where the last package added is the first
    package checked for a resource. Each package can be any object
    that implements the AbstractPackage interface.
    """
    def __init__(self, package_list):
        super().__init__()
        self._package_list = package_list
        self._reset_caches()

    @classmethod
    def open(cls, filename):
        from .. import package
        # Metapackages are always read-only
        packages = []
        with open(filename, "r") as f:
            for name in f.readlines():
                name = name.strip("\uFEFF\n")
                packages.append(package.open_package(name, mode="r"))
        return cls(packages)

    def scan_index(self, filter=None):
        if filter is None:
            return self._entry_cache.keys()
        return (rid for rid in self._entry_cache
                    if filter.match(rid))

    def __getitem__(self, key):
        return self._entry_cache[key]

    def _get_content(self, resource):
        # This should never actually get called as we shouldn't be in
        # the package field of any resources. Still, if somebody
        # *does* decide to call this method directly, it should work.
        return _resource.package._get_content(resource)

    def flush_index_cache(self):
        self._entry_cache = None

    def _reset_caches(self):
        self._entry_cache = {}
        for package in self._package_list:
            for id in package.scan_index():
                self._entry_cache[id] = package[id]
            package.flush_index_cache()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['s4sdk', 's4sdk.package', 's4sdk.resource']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'argparse>=1.4.0,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'pandas>=1.4.3,<2.0.0']

setup_kwargs = {
    'name': 's4sdk',
    'version': '0.1.1a2',
    'description': '',
    'long_description': None,
    'author': 'Roy Chotechuang',
    'author_email': 'htraexd@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
